/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2013, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPrediction.h
    \brief    prediction class (header)
*/

#ifndef __TCOMPREDICTION__
#define __TCOMPREDICTION__


// Include files
#include "TComPic.h"
#include "TComMotionInfo.h"
#include "TComPattern.h"
#include "TComTrQuant.h"
#include "TComInterpolationFilter.h"
#include "TComWeightPrediction.h"

#if SVC_UPSAMPLING
#include "TComDataCU.h"
#include "TComUpsampleFilter.h"
#endif
//! \ingroup TLibCommon
//! \{

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// prediction class
class TComPrediction : public TComWeightPrediction
{
protected:
  Int*      m_piYuvExt;
  Int       m_iYuvExtStride;
  Int       m_iYuvExtHeight;
  
  TComYuv   m_acYuvPred[2];
  TComYuv   m_cYuvPredTemp;
  TComYuv m_filteredBlock[4][4];
  TComYuv m_filteredBlockTmp[4];
  
  TComInterpolationFilter m_if;
  
#if SVC_UPSAMPLING
  TComUpsampleFilter      m_cUsf;
#endif

  Pel*   m_pLumaRecBuffer;       ///< array for downsampled reconstructed luma sample 
  Int    m_iLumaRecStride;       ///< stride of #m_pLumaRecBuffer array

  Void xPredIntraAng            (Int bitDepth, Int* pSrc, Int srcStride, Pel*& rpDst, Int dstStride, UInt width, UInt height, UInt dirMode, Bool blkAboveAvailable, Bool blkLeftAvailable, Bool bFilter );
  Void xPredIntraPlanar         ( Int* pSrc, Int srcStride, Pel* rpDst, Int dstStride, UInt width, UInt height );
  
  // motion compensation functions
#if DEC_EILR_BLK
  Void xPredInterBlkILR  ( TComPic *pEILRPic, TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvPred, Bool bi
#if ILR_TM_ENH_UCM
                           ,TComPicSym* pStoredPicSym
#endif
    );
  Void xPredInterUniEL   ( TComPic *pEILRPic, TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Bool bi=false
#if ILR_TM_ENH_UCM
                          ,TComPicSym* pStoredPicSym=NULL
#endif
    );
  Void xPredInterBiEL    ( TComPic *pEILRPic, TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvPred
#if ILR_TM_ENH_UCM
                          ,TComPicSym* pStoredPicSym
#endif
    );  
#endif
  Void xPredInterUni            ( TComDataCU* pcCU,                          UInt uiPartAddr,               Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Bool bi=false          );
  Void xPredInterBi             ( TComDataCU* pcCU,                          UInt uiPartAddr,               Int iWidth, Int iHeight,                         TComYuv*& rpcYuvPred );
  Void xPredInterLumaBlk  ( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi );
  Void xPredInterChromaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi );
  Void xWeightedAverage         ( TComYuv* pcYuvSrc0, TComYuv* pcYuvSrc1, Int iRefIdx0, Int iRefIdx1, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvDst );
  
  Void xGetLLSPrediction ( TComPattern* pcPattern, Int* pSrc0, Int iSrcStride, Pel* pDst0, Int iDstStride, UInt uiWidth, UInt uiHeight, UInt uiExt0 );

  Void xDCPredFiltering( Int* pSrc, Int iSrcStride, Pel*& rpDst, Int iDstStride, Int iWidth, Int iHeight );
  Bool xCheckIdenticalMotion    ( TComDataCU* pcCU, UInt PartAddr);
#if ILR_TM_ENH
  TComYuv *m_pEnhPred[2];  //0: Up(BLx); 1: MC(ELx);
  Int     m_iSRNum[2];  // 2/3/4
#if ILR_TM_ENH_EXPLICIT_MODE
  Char    **m_pchEILRBlockMode;
#endif
#if DEC_EILR_BLK
  TComYuv   *m_acYuvPred4ELR[2];
#endif
#endif

public:
  TComPrediction();
  virtual ~TComPrediction();
#if ILR_TM_ENH  
  Void    initTempBuff(Bool bEnhPred=false, Bool bEnhPred4BlkDec=false);
  Void    setSR(Int *piSR) { *m_iSRNum = *piSR; *(m_iSRNum+1) = *(piSR+1); }
#else
  Void    initTempBuff();
#endif  
#if INTRA_BL
  Void getBaseBlk ( TComDataCU* pcCU, TComYuv* pcYuvPred, Int iPartAddr, Int iWidth, Int iHeight );
#endif
  // inter
  Void motionCompensation         ( TComDataCU*  pcCU, TComYuv* pcYuvPred, RefPicList eRefPicList = REF_PIC_LIST_X, Int iPartIdx = -1 );
#if DEC_EILR_BLK
  Void motionCompensationEL( TComDataCU*  pcCU, TComYuv* pcYuvPred, TComPic *pEILRPic
#if ILR_TM_ENH_UCM
    ,TComPicSym* pStoredPicSym
#endif
    );
#endif

  // motion vector prediction
  Void getMvPredAMVP              ( TComDataCU* pcCU, UInt uiPartIdx, UInt uiPartAddr, RefPicList eRefPicList, TComMv& rcMvPred );
  
  // Angular Intra
  Void predIntraLumaAng           ( TComPattern* pcTComPattern, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Bool bAbove, Bool bLeft );
  Void predIntraChromaAng         ( Int* piSrc, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Bool bAbove, Bool bLeft );
  
  Pel  predIntraGetPredValDC      ( Int* pSrc, Int iSrcStride, UInt iWidth, UInt iHeight, Bool bAbove, Bool bLeft );
  
  Int* getPredicBuf()             { return m_piYuvExt;      }
  Int  getPredicBufWidth()        { return m_iYuvExtStride; }
  Int  getPredicBufHeight()       { return m_iYuvExtHeight; }

#if SVC_UPSAMPLING
#if SCALED_REF_LAYER_OFFSETS
  Void upsampleBasePic( TComPicYuv* pcUsPic, TComPicYuv* pcBasePic, TComPicYuv* pcTempPic, const Window window );
#else
  Void upsampleBasePic( TComPicYuv* pcUsPic, TComPicYuv* pcBasePic, TComPicYuv* pcTempPic );
#endif
#endif

#if ILR_TM_ENH
  Void getProcessUnitSize(Int &iPUWidth, Int &iPUHeight)
  {
  #if ILR_TM_ENH_UCM
  #if ILR_TM_ENH_UCM==3
    iPUWidth = 16;
    iPUHeight = 16;
  #else
    iPUWidth = (g_uiMaxCUWidth>>g_uiMaxCUDepth)*m_iSRNum[0];
    iPUHeight = (g_uiMaxCUHeight>>g_uiMaxCUDepth)*m_iSRNum[1];
  #endif
  #else
    iPUWidth = (m_iSRNum[0]<<3)&(~0xf); //16; 
    iPUHeight = (m_iSRNum[1]<<3)&(~0xf); //16; 
  #endif
  }

  Void setRefAndMv(TComDataCU *pcCU, UInt uiAbsPartIdx, Int *piRefIdx, TComMv *pcMv)
  {
    piRefIdx[0] = pcCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(uiAbsPartIdx);
    piRefIdx[1] = pcCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(uiAbsPartIdx);  
    if(piRefIdx[0]>=0)
      pcMv[0] = pcCU->getCUMvField( REF_PIC_LIST_0 )->getMv( uiAbsPartIdx );
    if(piRefIdx[1]>=0)
      pcMv[1] = pcCU->getCUMvField( REF_PIC_LIST_1 )->getMv( uiAbsPartIdx );
  }
  
  Void derivePartSize(Int iRefIdx[4][2], TComMv cMvArr[4][2], Int &iPartNum, PartSize &ePartSize)
  {
    iPartNum=4;
    ePartSize = SIZE_NxN;
    if(    ((iRefIdx[0][0] == iRefIdx[1][0] && cMvArr[0][0]==cMvArr[1][0]) && (iRefIdx[0][1] == iRefIdx[1][1] && cMvArr[0][1]==cMvArr[1][1]))
        && ((iRefIdx[0][0] == iRefIdx[2][0] && cMvArr[0][0]==cMvArr[2][0]) && (iRefIdx[0][1] == iRefIdx[2][1] && cMvArr[0][1]==cMvArr[2][1]))
        && ((iRefIdx[0][0] == iRefIdx[3][0] && cMvArr[0][0]==cMvArr[3][0]) && (iRefIdx[0][1] == iRefIdx[3][1] && cMvArr[0][1]==cMvArr[3][1]))
      )
    {
      ePartSize = SIZE_2Nx2N;
      iPartNum = 1;
    }
    else if(   ((iRefIdx[0][0] == iRefIdx[1][0] && cMvArr[0][0]==cMvArr[1][0]) && (iRefIdx[0][1] == iRefIdx[1][1] && cMvArr[0][1]==cMvArr[1][1]))
            && ((iRefIdx[2][0] == iRefIdx[3][0] && cMvArr[2][0]==cMvArr[3][0]) && (iRefIdx[2][1] == iRefIdx[3][1] && cMvArr[2][1]==cMvArr[3][1])) )
    {
      ePartSize = SIZE_2NxN;
      iPartNum = 2;       
    }
    else if(   ((iRefIdx[0][0] == iRefIdx[2][0] && cMvArr[0][0]==cMvArr[2][0]) && (iRefIdx[0][1] == iRefIdx[2][1] && cMvArr[0][1]==cMvArr[2][1]))
            && ((iRefIdx[1][0] == iRefIdx[3][0] && cMvArr[1][0]==cMvArr[3][0]) && (iRefIdx[1][1] == iRefIdx[3][1] && cMvArr[1][1]==cMvArr[3][1])) )
    {
      ePartSize = SIZE_Nx2N;
      iPartNum = 2;
    }

  }
  

  //Void calcEnhBlk(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight); 

  Void xPredInterLumaBlkByPosBilinear(TComPicYuv *refPic, Int iBlkX, Int iBlkY,  TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi );
  Void xPredInterChromaBlkByPosBilinear(TComPicYuv *refPic, Int iBlkX, Int iBlkY, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi );

#if ILR_TM_ENH_WEIGHT_SIGNAL
  Void enhPicWithTMWeight(TComPic *pcPicSrc, TComPic *pcPicDst
#if ILR_TM_ENH_UCM
    , TComPicSym *pcPicSym
#endif
    );
  Void enhPicWithTMWeightDiff(TComPic *pcPicSrc, TComPic *pcPicDst
#if ILR_TM_ENH_UCM
    , TComPicSym *pcPicSym
#endif
    );
  Void getEnhBlockMPUWeight(TComPic *pcPicSrc, TComPic *pcPicDst, TComMv *pcBlkPos, TComMv *pcBlkSize
#if ILR_TM_ENH_UCM
    , TComPicSym *pcPicSymBL
#endif
    );
#if DEC_EILR_BLK
  Void getEnhBlockMPUWeightDec(TComPic *pcPicSrc, TComPic *pcEILRPic, TComYuv *&pcPicDst, UInt uiPartAddr, TComMv *pcStartPos, TComMv *pcBlkPos, TComMv *pcBlkSize
#if ILR_TM_ENH_UCM
    , TComPicSym *pcPicSymBL
#endif
    );    
  Void calcEnhBlkModeDiffDec(TComPicYuv *pcPicYuvSrc, TComYuv *&pcPicYuvDst, UInt uiPartAddr, TComMv *pSPos, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iWeight
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
                             , Int iOffset
#endif
    );
  Void calcEnhBlkOneCompDec(TComPicYuv *pcPicYuvSrc, TComYuv *&pcPicYuvDst, UInt uiPartAddr, TComMv *pSPos, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iCompIdx);
#endif
  Void getEnhBlockMPUWeight_D(TComPic *pcPicSrc, TComPic *pcPicDst, TComMv *pcBlkPos, TComMv *pcBlkSize
#if ILR_TM_ENH_UCM
    , TComPicSym *pcPicSymBL
#endif
    );
  Void calcEnhBlkModeDiff(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iWeight
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
                          ,Int iOffset
#endif
    );

  //with diff/res signal is ready;
  //Void calcEnhBlk2(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight);
  Void calcEnhBlkModeDiff2(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iWeight
#if (ILR_TM_ENH_WEIGHT_SIGNAL==2)
                           ,Int iOffset
#endif
                          );
#else
  Void enhPicWithTM(TComPic *pcPicSrc, TComPic *pcPicDst
#if ILR_TM_ENH_UCM
    , TComPicSym *pcPicSym
#endif
    );
  Void getEnhBlockMPU(TComPic *pcPicSrc, TComPic *pcPicDst, TComMv *pcBlkPos, TComMv *pcBlkSize
#if ILR_TM_ENH_UCM
    , TComPicSym *pcPicSym
#endif
    );
#endif

#if ILR_TM_ENH_EXPLICIT_MODE
  Void allocateEILRBlockMode(UInt uiSWidth, UInt uiSHeight, UInt uiMinPUWidth=4, UInt uiMinPUHeight=4);
  Char**  getEILRBlockMode() { return m_pchEILRBlockMode; }
  Void copyBlkOneComp(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iCompIdx);
  Void calcEnhBlkOneComp(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iCompIdx);
  //encoder side;
  Void calcEnhBlk2OneComp(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iCompIdx);
#endif

#endif  //ILR_TM_ENH

};

//! \}

#endif // __TCOMPREDICTION__
